/**
 * Copyright (c) 2014 Zeligsoft, CEA and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package org.eclipse.papyrus.qompass.modellibs.core.xtend;

import com.google.common.base.Objects;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.papyrus.qompass.designer.core.EnumService;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.BehaviorUtil;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.GlobalConstants;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.StateMachineUtil;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.UMLAction;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.UMLChain;
import org.eclipse.uml2.uml.Behavior;
import org.eclipse.uml2.uml.BehavioralFeature;
import org.eclipse.uml2.uml.CallEvent;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Constraint;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Enumeration;
import org.eclipse.uml2.uml.EnumerationLiteral;
import org.eclipse.uml2.uml.Event;
import org.eclipse.uml2.uml.Interface;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.OpaqueBehavior;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.PackageableElement;
import org.eclipse.uml2.uml.Parameter;
import org.eclipse.uml2.uml.Port;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Pseudostate;
import org.eclipse.uml2.uml.PseudostateKind;
import org.eclipse.uml2.uml.Region;
import org.eclipse.uml2.uml.State;
import org.eclipse.uml2.uml.StateMachine;
import org.eclipse.uml2.uml.Stereotype;
import org.eclipse.uml2.uml.Transition;
import org.eclipse.uml2.uml.Trigger;
import org.eclipse.uml2.uml.Type;
import org.eclipse.uml2.uml.UMLPackage;
import org.eclipse.uml2.uml.Vertex;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;

/**
 * This class contains the transformation from flat UML-RT state machines to the
 * C/C++ language model.
 * 
 * It implements the algorithms described in the technical report
 * 
 * E. Posse. "Transforming flat UML-RT State Machines to a C/C++ language model".
 * Technical Report ZTR-2014-EP-002, Version 2, Zeligsoft, Sep 2014.
 * 
 * Ansgar Radermacher (CEA): translation to UML transformation
 * @author eposse
 */
@SuppressWarnings("all")
public class UMLFlatModel2Cpp {
  /**
   * The source state machine to transform
   */
  private StateMachine machine;
  
  private org.eclipse.uml2.uml.Class clazz;
  
  /**
   * Elements that go into the generated model.
   */
  private Enumeration statesDeclaration;
  
  private Property currentStateField;
  
  private Map<State, EnumerationLiteral> stateEnumerators;
  
  private Property historyTableDeclaration;
  
  private Operation saveHistoryFunction;
  
  private Map<Transition, OpaqueBehavior> actionChainFunctions;
  
  private Map<Pseudostate, OpaqueBehavior> junctionPointFunctions;
  
  private Map<Pseudostate, OpaqueBehavior> choicePointFunctions;
  
  private Map<State, OpaqueBehavior> stateFunctions;
  
  private Operation injectFunc;
  
  private Operation initializeFunc;
  
  private Stereotype refStereo;
  
  private Stereotype constStereo;
  
  public UMLFlatModel2Cpp() {
    HashMap<State, EnumerationLiteral> _newHashMap = CollectionLiterals.<State, EnumerationLiteral>newHashMap();
    this.stateEnumerators = _newHashMap;
    HashMap<Transition, OpaqueBehavior> _newHashMap_1 = CollectionLiterals.<Transition, OpaqueBehavior>newHashMap();
    this.actionChainFunctions = _newHashMap_1;
    HashMap<Pseudostate, OpaqueBehavior> _newHashMap_2 = CollectionLiterals.<Pseudostate, OpaqueBehavior>newHashMap();
    this.junctionPointFunctions = _newHashMap_2;
    HashMap<Pseudostate, OpaqueBehavior> _newHashMap_3 = CollectionLiterals.<Pseudostate, OpaqueBehavior>newHashMap();
    this.choicePointFunctions = _newHashMap_3;
    HashMap<State, OpaqueBehavior> _newHashMap_4 = CollectionLiterals.<State, OpaqueBehavior>newHashMap();
    this.stateFunctions = _newHashMap_4;
  }
  
  /**
   * This is the main method of the transformation. It performs the
   * transformation by invoking methods that generate each part of the
   * target language model.
   */
  public void transform(final StateMachine m) {
    this.machine = m;
    Element _owner = m.getOwner();
    this.clazz = ((org.eclipse.uml2.uml.Class) _owner);
    this.generateStatesDeclaration();
    this.generateCurrentStateField();
    this.generateHistoryTableDeclaration();
    this.generateSaveHistoryFunction();
    this.generateAllActionChainFunctions();
    this.generateAllJunctionFunctions();
    this.generateAllChoicePointFunctions();
    this.generateAllStateFunctions();
    this.generateInjectFunc();
    this.generateInitializeFunc();
  }
  
  /**
   * Builds an enum type for the states of the state machine.
   * 
   * The generated code would be something like:
   * 
   * <p><pre>
   * <code>enum State { s0, s1, s1_s0, s1_s1, s2, ... };</code>
   * </pre>
   */
  public EnumerationLiteral generateStatesDeclaration() {
    EnumerationLiteral _xblockexpression = null;
    {
      Element _owner = this.machine.getOwner();
      final org.eclipse.uml2.uml.Class clazz = ((org.eclipse.uml2.uml.Class) _owner);
      EClass _enumeration = UMLPackage.eINSTANCE.getEnumeration();
      Classifier _createNestedClassifier = clazz.createNestedClassifier(GlobalConstants.STATE_TYPE_NAME, _enumeration);
      this.statesDeclaration = ((Enumeration) _createNestedClassifier);
      EList<State> _states = StateMachineUtil.states(this.machine);
      for (final State s : _states) {
        {
          String _name = s.getName();
          boolean _equals = Objects.equal(_name, null);
          if (_equals) {
            s.setName("undefined");
          }
          String _name_1 = s.getName();
          final EnumerationLiteral stateLiteral = this.statesDeclaration.createOwnedLiteral(_name_1);
          this.stateEnumerators.put(s, stateLiteral);
        }
      }
      _xblockexpression = this.statesDeclaration.createOwnedLiteral(GlobalConstants.UNVISITED);
    }
    return _xblockexpression;
  }
  
  /**
   * Generates a field to hold the current state.
   * 
   * The generated code would be something like:
   * --
   * <p><pre>
   * <code>
   * State currentState;
   * </code>
   * </pre>
   */
  public Property generateCurrentStateField() {
    Property _createOwnedAttribute = this.clazz.createOwnedAttribute(GlobalConstants.CURRENT_STATE_FIELD_NAME, this.statesDeclaration);
    return this.currentStateField = _createOwnedAttribute;
  }
  
  /**
   * Generates a declaration for the history table for the state machine.
   * 
   * The generated code would be something like:
   * 
   * <p>
   * <code>State[] history = { UNDEFINED, ..., UNDEFINED };
   */
  public void generateHistoryTableDeclaration() {
    Property _createOwnedAttribute = this.clazz.createOwnedAttribute(GlobalConstants.HISTORY_TABLE_NAME, this.statesDeclaration);
    this.historyTableDeclaration = _createOwnedAttribute;
    EList<State> _states = StateMachineUtil.states(this.machine);
    final int numStates = _states.size();
    String _name = this.clazz.getName();
    Operation ctor = this.clazz.getOwnedOperation(_name, null, null);
    OpaqueBehavior ctorOB = null;
    boolean _notEquals = (!Objects.equal(ctor, null));
    if (_notEquals) {
      EList<Behavior> _methods = ctor.getMethods();
      Behavior _get = _methods.get(0);
      ctorOB = ((OpaqueBehavior) _get);
    }
    boolean _equals = Objects.equal(ctor, null);
    if (_equals) {
      String _name_1 = this.clazz.getName();
      Operation _createOperation = BehaviorUtil.createOperation(this.clazz, _name_1, null);
      ctor = _createOperation;
      OpaqueBehavior _createOpaqueBehavior = BehaviorUtil.createOpaqueBehavior(this.clazz, ctor);
      ctorOB = _createOpaqueBehavior;
    }
    StringConcatenation _builder = new StringConcatenation();
    _builder.append("int i = 0;");
    _builder.newLine();
    _builder.append("while (i<");
    _builder.append(numStates, "");
    _builder.append(") {");
    _builder.newLineIfNotEmpty();
    _builder.append("\t");
    _builder.append(GlobalConstants.HISTORY_TABLE_NAME, "\t");
    _builder.append("[i++] = ");
    _builder.append(GlobalConstants.UNVISITED, "\t");
    _builder.append(";");
    _builder.newLineIfNotEmpty();
    _builder.append("}");
    BehaviorUtil.set(ctorOB, _builder.toString());
  }
  
  /**
   * Generates a function that saves history.
   * 
   * The code generated is as follows:
   * 
   * <p><pre>
   * <code>
   * void saveHistory(State compositeState, State subState) {
   *     history[compositeState] = subState;
   * }
   * <code>
   * </pre>
   * 
   * where <code>State</code> is the capsule's state type (an enum) and
   * <code>history</code> is the capsule's history table.
   * 
   * <p><b>Note:</b> The current implementation generates this as a normal
   * function but it should be either a macro or an inline function.
   * However the C/C++ language model does not currently support these.
   * 
   * @see
   *  #generateStatesDeclaration
   *  #generateHistoryTableDeclaration
   */
  public void generateSaveHistoryFunction() {
    Operation _createOwnedOperation = this.clazz.createOwnedOperation(GlobalConstants.SAVE_HISTORY_FUNC_NAME, null, null);
    this.saveHistoryFunction = _createOwnedOperation;
    Enumeration _stateType = this.getStateType();
    this.saveHistoryFunction.createOwnedParameter("compositeState", _stateType);
    Enumeration _stateType_1 = this.getStateType();
    this.saveHistoryFunction.createOwnedParameter("subState", _stateType_1);
    Behavior funcOB = this.clazz.getOwnedBehavior(GlobalConstants.SAVE_HISTORY_FUNC_NAME);
    boolean _equals = Objects.equal(funcOB, null);
    if (_equals) {
      OpaqueBehavior _createOpaqueBehavior = BehaviorUtil.createOpaqueBehavior(this.clazz, this.saveHistoryFunction);
      funcOB = _createOpaqueBehavior;
    } else {
      funcOB.setSpecification(this.saveHistoryFunction);
    }
    BehaviorUtil.set(funcOB, "history[compositeState] = subState;");
  }
  
  /**
   * Generate function declarations for transition action chains.
   * 
   * <p>Each function generated will have a sequence of calls, invoking either
   * the action functions generated by {@link generateActionFunc} for
   * transition, state entry and state exit actions, as well as invoking
   * "save history" actions produced by the flattening transformation.
   */
  public void generateAllActionChainFunctions() {
    Region _region = StateMachineUtil.region(this.machine);
    EList<Transition> _transitions = _region.getTransitions();
    for (final Transition t : _transitions) {
      {
        final OpaqueBehavior f = this.generateActionFunc(t);
        this.actionChainFunctions.put(t, f);
      }
    }
  }
  
  /**
   * Generates the function declaration for a single action chain.
   */
  public OpaqueBehavior generateActionFunc(final Transition t) {
    String _name = t.getName();
    final String transitionFuncName = ((GlobalConstants.TRANS_ACTION_FUNC_PREFIX + "__") + _name);
    final Operation transitionFunc = this.clazz.createOwnedOperation(transitionFuncName, null, null);
    this.createRTMessageParam(transitionFunc, GlobalConstants.CHAIN_FUNC_PARAM);
    Behavior transitionFuncOB = t.getEffect();
    boolean _equals = Objects.equal(transitionFuncOB, null);
    if (_equals) {
      OpaqueBehavior _createOpaqueBehavior = BehaviorUtil.createOpaqueBehavior(this.clazz, transitionFunc);
      transitionFuncOB = _createOpaqueBehavior;
    } else {
      transitionFuncOB.setSpecification(transitionFunc);
      EList<Behavior> _ownedBehaviors = this.clazz.getOwnedBehaviors();
      _ownedBehaviors.add(transitionFuncOB);
    }
    String _name_1 = t.getName();
    final String chainFuncName = ((GlobalConstants.ACTION_FUNC_PREFIX + "__") + _name_1);
    final Operation chainFunc = this.clazz.createOwnedOperation(chainFuncName, null, null);
    this.createRTMessageParam(chainFunc, GlobalConstants.CHAIN_FUNC_PARAM);
    final OpaqueBehavior chainFuncOB = BehaviorUtil.createOpaqueBehavior(this.clazz, chainFunc);
    String code = "";
    UMLChain _chain = UMLChain.chain(t);
    List<UMLAction> _actions = _chain.getActions();
    for (final UMLAction action : _actions) {
      String _code = code;
      BehavioralFeature _specification = action.getSpecification();
      CharSequence _call = this.call(_specification, action.callParams);
      code = (_code + _call);
    }
    BehaviorUtil.set(chainFuncOB, code);
    return ((OpaqueBehavior) chainFuncOB);
  }
  
  /**
   * Generates all functions corresponding to choice points.
   */
  public void generateAllChoicePointFunctions() {
    Region _region = StateMachineUtil.region(this.machine);
    EList<Pseudostate> _choicePoints = StateMachineUtil.choicePoints(_region);
    for (final Pseudostate c : _choicePoints) {
      {
        final OpaqueBehavior f = this.generateChoiceFunc(c);
        this.choicePointFunctions.put(c, f);
      }
    }
  }
  
  /**
   * Generates the function corresponding to a given choice point.
   */
  public OpaqueBehavior generateChoiceFunc(final Pseudostate p) {
    String _name = p.getName();
    final String funcName = (GlobalConstants.CHOICE_FUNC_PREFIX + _name);
    final Operation funcDecl = BehaviorUtil.createOperation(this.clazz, funcName, null);
    final Parameter param = this.createRTMessageParam(funcDecl, GlobalConstants.CHOICE_FUNC_PARAM);
    final OpaqueBehavior funcDeclOB = BehaviorUtil.createOpaqueBehavior(this.clazz, funcDecl);
    String code = "";
    EList<Transition> _outgoings = p.getOutgoings();
    for (final Transition t : _outgoings) {
      String _code = code;
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("if (");
      Constraint _guard = t.getGuard();
      String _body = BehaviorUtil.body(_guard);
      _builder.append(_body, "");
      _builder.append(") {");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      OpaqueBehavior _get = this.actionChainFunctions.get(t);
      String _name_1 = param.getName();
      CharSequence _call = this.call(_get, _name_1);
      _builder.append(_call, "\t");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      _builder.append("// TODO: return instead");
      _builder.newLine();
      _builder.append("\t");
      String _name_2 = this.currentStateField.getName();
      _builder.append(_name_2, "\t");
      _builder.append(" = ");
      String _destination = this.getDestination(t, false);
      _builder.append(_destination, "\t");
      _builder.append(";");
      _builder.newLineIfNotEmpty();
      _builder.append("}");
      _builder.newLine();
      code = (_code + _builder);
    }
    BehaviorUtil.set(funcDeclOB, code);
    return funcDeclOB;
  }
  
  /**
   * Generates all functions corresponding to junction points.
   */
  public void generateAllJunctionFunctions() {
    Region _region = StateMachineUtil.region(this.machine);
    EList<Pseudostate> _junctionPoints = StateMachineUtil.junctionPoints(_region);
    for (final Pseudostate j : _junctionPoints) {
      {
        final OpaqueBehavior f = this.generateJunctionFunc(j);
        this.junctionPointFunctions.put(j, f);
      }
    }
  }
  
  /**
   * Generates the function corresponding to a specific junction point.
   */
  public OpaqueBehavior generateJunctionFunc(final Pseudostate p) {
    OpaqueBehavior _xblockexpression = null;
    {
      String _name = p.getName();
      final String funcName = (GlobalConstants.JUNCTION_FUNC_PREFIX + _name);
      final Operation funcDecl = BehaviorUtil.createOperation(this.clazz, funcName, null);
      final OpaqueBehavior funcDeclOB = BehaviorUtil.createOpaqueBehavior(this.clazz, funcDecl);
      this.createRTMessageParam(funcDecl, GlobalConstants.JUNC_FUNC_PARAM);
      EList<Transition> _outgoings = p.getOutgoings();
      final Transition t = _outgoings.get(0);
      String code = "";
      boolean _notEquals = (!Objects.equal(t, null));
      if (_notEquals) {
        final OpaqueBehavior actFunc = this.actionChainFunctions.get(t);
        String _code = code;
        CharSequence _call = this.call(actFunc, GlobalConstants.JUNC_FUNC_PARAM);
        code = (_code + _call);
      }
      String _code_1 = code;
      String _destination = this.getDestination(t, false);
      code = (_code_1 + _destination);
      BehaviorUtil.set(funcDeclOB, code);
      _xblockexpression = funcDeclOB;
    }
    return _xblockexpression;
  }
  
  /**
   * Generates all functions corresponding to states.
   */
  public void generateAllStateFunctions() {
    EList<State> _states = StateMachineUtil.states(this.machine);
    for (final State s : _states) {
      {
        final OpaqueBehavior f = this.generateStateFunc(s);
        this.stateFunctions.put(s, f);
      }
    }
  }
  
  /**
   * Generates the function corresponding to a given state.
   */
  public OpaqueBehavior generateStateFunc(final State state) {
    OpaqueBehavior _xblockexpression = null;
    {
      String _name = state.getName();
      final String name = (GlobalConstants.STATE_FUNC_PREFIX + _name);
      final Enumeration retType = this.statesDeclaration;
      final Operation func = BehaviorUtil.createOperation(this.clazz, name, retType);
      final Parameter param = this.createRTMessageParam(func, GlobalConstants.STATE_FUNC_PARAM);
      final OpaqueBehavior funcOB = BehaviorUtil.createOpaqueBehavior(this.clazz, func);
      final HashMap<Port, HashSet<Transition>> table = this.getPortTransitionsTable(state);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("switch (");
      CharSequence _portCond = this.getPortCond(param);
      _builder.append(_portCond, "");
      _builder.append(") {");
      _builder.newLineIfNotEmpty();
      {
        Set<Port> _keySet = table.keySet();
        for(final Port port : _keySet) {
          _builder.append("\t");
          _builder.append("case ");
          CharSequence _enumeratorFor = this.enumeratorFor(port);
          _builder.append(_enumeratorFor, "\t");
          _builder.append(":");
          _builder.newLineIfNotEmpty();
          _builder.append("\t");
          _builder.append("\t");
          _builder.append("switch(");
          CharSequence _sigCond = this.getSigCond(param);
          _builder.append(_sigCond, "\t\t");
          _builder.append(") {");
          _builder.newLineIfNotEmpty();
          {
            HashSet<Transition> _get = table.get(port);
            for(final Transition t : _get) {
              {
                EList<Trigger> _triggers = t.getTriggers();
                final Function1<Trigger, Boolean> _function = new Function1<Trigger, Boolean>() {
                  public Boolean apply(final Trigger it) {
                    EList<Port> _ports = it.getPorts();
                    return Boolean.valueOf(_ports.contains(port));
                  }
                };
                Iterable<Trigger> _filter = IterableExtensions.<Trigger>filter(_triggers, _function);
                for(final Trigger trigger : _filter) {
                  _builder.append("\t");
                  _builder.append("\t");
                  _builder.append("case ");
                  Event _event = trigger.getEvent();
                  CharSequence _enumeratorFor_1 = this.enumeratorFor(((CallEvent) _event));
                  _builder.append(_enumeratorFor_1, "\t\t");
                  _builder.append(":");
                  _builder.newLineIfNotEmpty();
                  _builder.append("\t");
                  _builder.append("\t");
                  _builder.append("\t");
                  {
                    Constraint _guard = t.getGuard();
                    boolean _notEquals = (!Objects.equal(_guard, null));
                    if (_notEquals) {
                      _builder.append("if (t.guard) { ");
                    }
                  }
                  _builder.newLineIfNotEmpty();
                  _builder.append("\t");
                  _builder.append("\t");
                  _builder.append("\t");
                  OpaqueBehavior _get_1 = this.actionChainFunctions.get(t);
                  String _name_1 = _get_1.getName();
                  _builder.append(_name_1, "\t\t\t");
                  _builder.append("(");
                  String _name_2 = param.getName();
                  _builder.append(_name_2, "\t\t\t");
                  _builder.append(");");
                  _builder.newLineIfNotEmpty();
                  _builder.append("\t");
                  _builder.append("\t");
                  _builder.append("\t");
                  _builder.append("return ");
                  String _destination = this.getDestination(t, false);
                  _builder.append(_destination, "\t\t\t");
                  _builder.append(";");
                  _builder.newLineIfNotEmpty();
                  _builder.append("\t");
                  _builder.append("\t");
                  _builder.append("\t");
                  {
                    Constraint _guard_1 = t.getGuard();
                    boolean _notEquals_1 = (!Objects.equal(_guard_1, null));
                    if (_notEquals_1) {
                      _builder.append("}");
                    }
                  }
                  _builder.newLineIfNotEmpty();
                }
              }
            }
          }
          _builder.append("\t");
          _builder.append("\t");
          _builder.append("}");
          _builder.newLine();
        }
      }
      _builder.append("}");
      _builder.newLine();
      final String code = _builder.toString();
      BehaviorUtil.set(funcOB, code);
      _xblockexpression = funcOB;
    }
    return _xblockexpression;
  }
  
  /**
   * Builds a table that contains for each port, all the outgoing transitions
   * of a given state whose trigger has that port.
   * 
   * @param s     a {@link State}
   * @return a table T indexed by {@link Port}s such that for each port p, the
   *         entry T[p] contains all outgoing {@link Transition}s from s
   *         whose trigger has port p.
   */
  private HashMap<Port, HashSet<Transition>> getPortTransitionsTable(final State s) {
    final HashMap<Port, HashSet<Transition>> table = CollectionLiterals.<Port, HashSet<Transition>>newHashMap();
    EList<Transition> _outgoings = s.getOutgoings();
    for (final Transition t : _outgoings) {
      EList<Trigger> _triggers = t.getTriggers();
      for (final Trigger trigger : _triggers) {
        EList<Port> _ports = trigger.getPorts();
        for (final Port p : _ports) {
          {
            boolean _containsKey = table.containsKey(p);
            boolean _not = (!_containsKey);
            if (_not) {
              HashSet<Transition> _newHashSet = CollectionLiterals.<Transition>newHashSet();
              table.put(p, _newHashSet);
            }
            HashSet<Transition> _get = table.get(p);
            _get.add(t);
          }
        }
      }
    }
    return table;
  }
  
  /**
   * Obtains the function call corresponding to a transition's destination.
   * 
   * @param t - The {@link Transition}
   * @param init - Whether we are looking for the destination to be obtained
   *               in the ized method or in the inject method.
   */
  private String getDestination(final Transition t, final boolean init) {
    String _xblockexpression = null;
    {
      String retVal = null;
      Vertex _target = t.getTarget();
      if ((_target instanceof Pseudostate)) {
        Vertex _target_1 = t.getTarget();
        final Pseudostate ps = ((Pseudostate) _target_1);
        PseudostateKind _kind = ps.getKind();
        boolean _equals = Objects.equal(_kind, PseudostateKind.CHOICE_LITERAL);
        if (_equals) {
          final OpaqueBehavior func = this.choicePointFunctions.get(ps);
          String _name = func.getName();
          String _plus = (_name + "();");
          retVal = _plus;
        } else {
          PseudostateKind _kind_1 = ps.getKind();
          boolean _equals_1 = Objects.equal(_kind_1, PseudostateKind.JUNCTION_LITERAL);
          if (_equals_1) {
            final OpaqueBehavior func_1 = this.junctionPointFunctions.get(ps);
            String _name_1 = func_1.getName();
            String _plus_1 = (_name_1 + "();");
            retVal = _plus_1;
          }
        }
      } else {
        Vertex _target_2 = t.getTarget();
        EnumerationLiteral _get = this.stateEnumerators.get(_target_2);
        String _name_2 = _get.getName();
        retVal = _name_2;
      }
      if (init) {
        StringConcatenation _builder = new StringConcatenation();
        String _name_3 = this.currentStateField.getName();
        _builder.append(_name_3, "");
        _builder.append(" = ");
        _builder.append(retVal, "");
        _builder.append(";");
        retVal = _builder.toString();
      }
      _xblockexpression = retVal;
    }
    return _xblockexpression;
  }
  
  /**
   * Generates the main 'inject' function that receives and handles events.
   */
  public Operation generateInjectFunc() {
    Operation _xblockexpression = null;
    {
      Operation _createOperation = BehaviorUtil.createOperation(this.clazz, GlobalConstants.INJECT_FUNC_NAME, null);
      this.injectFunc = _createOperation;
      this.createRTMessageParam(this.injectFunc, GlobalConstants.INJECT_FUNC_PARAM);
      final OpaqueBehavior injectFuncOB = BehaviorUtil.createOpaqueBehavior(this.clazz, this.injectFunc);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("switch(stateCond) {");
      _builder.newLine();
      {
        EList<State> _states = StateMachineUtil.states(this.machine);
        for(final State s : _states) {
          _builder.append("case ");
          String _enumFor = this.enumFor(s);
          _builder.append(_enumFor, "");
          _builder.append(":");
          _builder.newLineIfNotEmpty();
          _builder.append("\t");
          String _name = this.currentStateField.getName();
          _builder.append(_name, "\t");
          _builder.append(" = ");
          OpaqueBehavior _get = this.stateFunctions.get(s);
          CharSequence _call = this.call(_get, GlobalConstants.INJECT_FUNC_PARAM);
          _builder.append(_call, "\t");
          _builder.newLineIfNotEmpty();
          _builder.append("\t");
          _builder.append("break;");
          _builder.newLine();
        }
      }
      _builder.append("}");
      _builder.newLine();
      final String code = _builder.toString();
      BehaviorUtil.set(injectFuncOB, code);
      _xblockexpression = this.injectFunc;
    }
    return _xblockexpression;
  }
  
  /**
   * Build the initialize function which performs the initial transition.
   * 
   * <p>This assumes that the top level of the state machine must have an
   * initial pseudo-state, and that there is exactly one outgoing transition
   * from such initial point.
   * 
   * <p> If there is no initial point, the body of the initialize method is
   * empty.
   */
  public void generateInitializeFunc() {
    Operation _createOperation = BehaviorUtil.createOperation(this.clazz, GlobalConstants.INITIALIZE_FUNC_NAME, null);
    this.initializeFunc = _createOperation;
    this.createRTMessageParam(this.initializeFunc, GlobalConstants.INITIALIZE_FUNC_PARAM);
    final OpaqueBehavior initializeFuncOB = BehaviorUtil.createOpaqueBehavior(this.clazz, this.initializeFunc);
    String code = "";
    Region _region = StateMachineUtil.region(this.machine);
    Pseudostate _initialState = StateMachineUtil.initialState(_region);
    boolean _notEquals = (!Objects.equal(_initialState, null));
    if (_notEquals) {
      Region _region_1 = StateMachineUtil.region(this.machine);
      Pseudostate _initialState_1 = StateMachineUtil.initialState(_region_1);
      EList<Transition> _outgoings = _initialState_1.getOutgoings();
      final Transition initialTransition = _outgoings.get(0);
      boolean _notEquals_1 = (!Objects.equal(initialTransition, null));
      if (_notEquals_1) {
        final OpaqueBehavior actFunc = this.actionChainFunctions.get(initialTransition);
        boolean _notEquals_2 = (!Objects.equal(actFunc, null));
        if (_notEquals_2) {
          String _code = code;
          CharSequence _call = this.call(actFunc, GlobalConstants.INITIALIZE_FUNC_PARAM);
          code = (_code + _call);
        }
      }
      String _code_1 = code;
      String _destination = this.getDestination(initialTransition, true);
      code = (_code_1 + _destination);
      BehaviorUtil.set(initializeFuncOB, code);
    }
  }
  
  public CharSequence call(final NamedElement fct, final String param) {
    StringConcatenation _builder = new StringConcatenation();
    String _name = fct.getName();
    _builder.append(_name, "");
    _builder.append("(");
    _builder.append(param, "");
    _builder.append(");");
    _builder.newLineIfNotEmpty();
    return _builder;
  }
  
  public CharSequence call(final NamedElement fct, final List<String> params) {
    StringConcatenation _builder = new StringConcatenation();
    {
      boolean _notEquals = (!Objects.equal(params, null));
      if (_notEquals) {
        String _name = fct.getName();
        _builder.append(_name, "");
        _builder.append("(");
        {
          boolean _hasElements = false;
          for(final String param : params) {
            if (!_hasElements) {
              _hasElements = true;
            } else {
              _builder.appendImmediate(", ", "");
            }
            _builder.append(param, "");
          }
        }
        _builder.append(");");
        _builder.newLineIfNotEmpty();
      } else {
        String _name_1 = fct.getName();
        _builder.append(_name_1, "");
        _builder.append("();");
        _builder.newLineIfNotEmpty();
      }
    }
    return _builder;
  }
  
  /**
   * Auxiliary methods
   */
  private Enumeration getStateType() {
    return this.statesDeclaration;
  }
  
  private Type getRTMessageType() {
    Type _xblockexpression = null;
    {
      Model _model = this.clazz.getModel();
      PackageableElement _packagedElement = _model.getPackagedElement("umlrts");
      final org.eclipse.uml2.uml.Package umlrts = ((org.eclipse.uml2.uml.Package) _packagedElement);
      PackageableElement _packagedElement_1 = umlrts.getPackagedElement("RTMessage");
      _xblockexpression = ((Type) _packagedElement_1);
    }
    return _xblockexpression;
  }
  
  private Parameter createRTMessageParam(final Operation operation, final String paramName) {
    Parameter _xblockexpression = null;
    {
      Type _rTMessageType = this.getRTMessageType();
      final Parameter param = operation.createOwnedParameter(paramName, _rTMessageType);
      boolean _equals = Objects.equal(this.refStereo, null);
      if (_equals) {
        Stereotype _applicableStereotype = param.getApplicableStereotype("C_Cpp::Ref");
        this.refStereo = _applicableStereotype;
      }
      boolean _notEquals = (!Objects.equal(this.refStereo, null));
      if (_notEquals) {
        param.applyStereotype(this.refStereo);
      }
      boolean _equals_1 = Objects.equal(this.constStereo, null);
      if (_equals_1) {
        Stereotype _applicableStereotype_1 = param.getApplicableStereotype("C_Cpp::Const");
        this.constStereo = _applicableStereotype_1;
      }
      boolean _notEquals_1 = (!Objects.equal(this.constStereo, null));
      if (_notEquals_1) {
        param.applyStereotype(this.constStereo);
      }
      _xblockexpression = param;
    }
    return _xblockexpression;
  }
  
  private Type getRTDataType(final Transition t) {
    Type _xblockexpression = null;
    {
      Type rtdataType = null;
      final Collection<Type> types = CollectionLiterals.<Type>newHashSet();
      EList<Trigger> _triggers = t.getTriggers();
      for (final Trigger trigger : _triggers) {
        {
          Event _event = trigger.getEvent();
          final CallEvent evt = ((CallEvent) _event);
          Operation _operation = evt.getOperation();
          final EList<Parameter> params = _operation.getOwnedParameters();
          boolean _isEmpty = params.isEmpty();
          boolean _not = (!_isEmpty);
          if (_not) {
            final Parameter first = params.get(0);
            Type _type = first.getType();
            types.add(_type);
          }
        }
      }
      int _size = types.size();
      boolean _equals = (_size == 1);
      if (_equals) {
        final Type type = ((Type[])Conversions.unwrapArray(types, Type.class))[0];
        rtdataType = type;
      }
      _xblockexpression = rtdataType;
    }
    return _xblockexpression;
  }
  
  private String enumFor(final State s) {
    EnumerationLiteral _get = this.stateEnumerators.get(s);
    return _get.getLabel();
  }
  
  private CharSequence _enumeratorFor(final Port port) {
    String _xblockexpression = null;
    {
      String _name = port.getName();
      final String enumerator = EnumService.literal("PortID", _name);
      _xblockexpression = enumerator;
    }
    return _xblockexpression;
  }
  
  private CharSequence _enumeratorFor(final CallEvent ce) {
    CharSequence _xblockexpression = null;
    {
      final Operation operation = ce.getOperation();
      Interface _interface = operation.getInterface();
      Element _owner = _interface.getOwner();
      String _name = ((NamedElement) _owner).getName();
      final String protocolCls = EnumService.literal("ProtocolClass", _name);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append(protocolCls, "");
      _builder.append("::signal_");
      String _name_1 = ce.getName();
      _builder.append(_name_1, "");
      _xblockexpression = _builder;
    }
    return _xblockexpression;
  }
  
  /**
   * Builds an expression to obtain the port enum id for the switch.
   * 
   * <p>It assumes that the message parameter to the inject function is a
   * pointer to the RTMessage type, as returned by {@link #getRTMessageType},
   * this is, the signature of the inject function must be:
   * 
   * <p><pre>
   * <code>void inject(UMLRTMessage * msg)</code>
   * </pre>
   * 
   * <p>It also assumes that the port id and signal id are accessible from
   * this type. Basically the assumption is that the relevant definitions are
   * as follows:
   * 
   * <p>
   * <pre>
   * <code>
   * class UMLRTMessage : ... {
   * public:
   *     UMLRTPort * destPort;
   *     UMLRTSignal * signal;
   * }
   * 
   * struct UMLRTPort {
   *     int id;
   *     // ...
   * }
   * 
   * class UMLRTSignal {
   * public:
   *     int id;
   *     // ...
   * }
   * </code>
   * </pre>
   * 
   * <p>... where the typed <code>UMLRTPortId</code> and
   * <code>UMLRTSignalId</code> can be cast to the <code>Port</code> and
   * <code>Signal</code> enum types generated in the state machine's class.
   * 
   * <p>Given this assumptions, the port condition generated has the form:
   * 
   * <p><pre><code>(Port)(msg->destPort)->id</code></pre>
   * 
   * <p>and the signal condition is:
   * 
   * <p><pre><code>(ProtocolX::Signal)(msg->signal)->getId()</code></pre>
   * 
   * <p>where <code>ProtocolX</code> is the name of the port's protocol
   */
  private CharSequence getPortCond(final Parameter param) {
    StringConcatenation _builder = new StringConcatenation();
    _builder.append("TODO");
    return _builder;
  }
  
  private CharSequence getSigCond(final Parameter param) {
    StringConcatenation _builder = new StringConcatenation();
    _builder.append("TODO");
    return _builder;
  }
  
  private CharSequence enumeratorFor(final Element port) {
    if (port instanceof Port) {
      return _enumeratorFor((Port)port);
    } else if (port instanceof CallEvent) {
      return _enumeratorFor((CallEvent)port);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(port).toString());
    }
  }
}
