/**
 * Copyright (c) 2014 Zeligsoft, CEA and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package org.eclipse.papyrus.qompass.modellibs.core.xtend;

import com.google.common.base.Objects;
import java.util.Collection;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.BehaviorUtil;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.GlobalConstants;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.StateMachineUtil;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.UMLAction;
import org.eclipse.papyrus.qompass.modellibs.core.xtend.UMLChain;
import org.eclipse.uml2.uml.Behavior;
import org.eclipse.uml2.uml.Constraint;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Event;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.OpaqueBehavior;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Pseudostate;
import org.eclipse.uml2.uml.PseudostateKind;
import org.eclipse.uml2.uml.Region;
import org.eclipse.uml2.uml.State;
import org.eclipse.uml2.uml.StateMachine;
import org.eclipse.uml2.uml.Transition;
import org.eclipse.uml2.uml.Trigger;
import org.eclipse.uml2.uml.UMLPackage;
import org.eclipse.uml2.uml.Vertex;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.Conversions;

/**
 * This class contains the transformation for flattening UML-RT state machines
 * 
 * It implements the algorithms described in the technical report
 * 
 * E. Posse. "Flattening UML-RT State Machines".
 * Technical Report ZTR-2014-EP-001, Version 2, Zeligsoft, Sep 2014.
 * 
 * @author eposse
 */
@SuppressWarnings("all")
public class UMLFlattener {
  private final String FRESH_NAME_PREFIX = "p";
  
  private final String FRESH_ENTRYPOINT_NAME_PREFIX = "entrypoint_";
  
  private final String FRESH_EXITPOINT_NAME_PREFIX = "exitpoint_";
  
  private final String FRESH_CHOICEPOINT_NAME_PREFIX = "c";
  
  private final String FRESH_JUNCTIONPOINT_NAME_PREFIX = "j";
  
  private String FRESH_TRANSITION_NAME_PREFIX = "transition_";
  
  private int freshNameCounter = 0;
  
  private int freshEntryPointNameCounter = 0;
  
  private int freshExitPointNameCounter = 0;
  
  private int freshChoicePointNameCounter = 0;
  
  private int freshJunctionPointNameCounter = 0;
  
  private int freshTransitionNameCounter = 0;
  
  private Behavior saveHistoryBehavior = null;
  
  private org.eclipse.uml2.uml.Class clazz;
  
  /**
   * Flattens a UML-RT state machine, removing composite states, and making
   * explicit transitions implied by group transitions, history points and
   * initial points.
   * 
   * This is the 'main' method of the transformer. It performs the
   * transformation in-place.
   * 
   * @param m a {@link StateMachine}
   * @return the flat {@link StateMachine} m with all composite states, group
   *          transitions, history and initial pseudo-states removed.
   */
  public StateMachine transform(final StateMachine m) {
    StateMachine intermediateModel = null;
    Element _owner = m.getOwner();
    this.clazz = ((org.eclipse.uml2.uml.Class) _owner);
    UMLChain.initChains();
    Region _region = StateMachineUtil.region(m);
    String _name = m.getName();
    this.phase0QualifyNames(_region, _name);
    this.phase1(m);
    StateMachine _phase2 = this.phase2(m);
    intermediateModel = _phase2;
    return intermediateModel;
  }
  
  /**
   * Change the names of states and transitions: prefix them with the names
   * of the stateMachine and composite states.
   */
  protected void phase0QualifyNames(final Region region, final String parentName) {
    EList<Transition> _transitions = region.getTransitions();
    for (final Transition transition : _transitions) {
      String _name = transition.getName();
      String _plus = ((parentName + "_") + _name);
      transition.setName(_plus);
    }
    EList<Vertex> _subvertices = region.getSubvertices();
    for (final Vertex vertex : _subvertices) {
      {
        String _name_1 = vertex.getName();
        String _plus_1 = ((parentName + "_") + _name_1);
        vertex.setName(_plus_1);
        if ((vertex instanceof State)) {
          final State state = ((State) vertex);
          EList<Region> _regions = state.getRegions();
          for (final Region subRegion : _regions) {
            String _name_2 = ((State)vertex).getName();
            this.phase0QualifyNames(subRegion, _name_2);
          }
        }
      }
    }
  }
  
  /**
   * Makes explicit transitions implied by group transitions, history points
   * and initial points.
   */
  protected StateMachine phase1(final StateMachine m) {
    EList<State> _states = StateMachineUtil.states(m);
    for (final State s : _states) {
      this.phase1ProcessState(s);
    }
    return m;
  }
  
  public void phase1ProcessState(final State state) {
    this.phase1ProcessSimpleState(state);
    EList<Region> _regions = state.getRegions();
    boolean _isEmpty = _regions.isEmpty();
    boolean _not = (!_isEmpty);
    if (_not) {
      this.phase1ProcessCompositeState(state);
    }
  }
  
  /**
   * Moves a simple state's entry actions to its incoming transitions and
   * the state's exit actions to its outgoing transitions.
   */
  public void phase1ProcessSimpleState(final State state) {
    Behavior _entry = state.getEntry();
    boolean _notEquals = (!Objects.equal(_entry, null));
    if (_notEquals) {
      String _entryFctName = this.entryFctName(state);
      final Operation entryOp = BehaviorUtil.createOperation(this.clazz, _entryFctName, null);
      Behavior _entry_1 = state.getEntry();
      String _entryFctName_1 = this.entryFctName(state);
      _entry_1.setName(_entryFctName_1);
      Behavior _entry_2 = state.getEntry();
      _entry_2.setSpecification(entryOp);
      EList<Transition> _incomings = state.getIncomings();
      for (final Transition t : _incomings) {
        UMLChain _chain = UMLChain.chain(t);
        Behavior _entry_3 = state.getEntry();
        _chain.append(_entry_3);
      }
    }
    Behavior _exit = state.getExit();
    boolean _notEquals_1 = (!Objects.equal(_exit, null));
    if (_notEquals_1) {
      String _exitFctName = this.exitFctName(state);
      final Operation exitOp = BehaviorUtil.createOperation(this.clazz, _exitFctName, null);
      Behavior _exit_1 = state.getExit();
      String _exitFctName_1 = this.exitFctName(state);
      _exit_1.setName(_exitFctName_1);
      Behavior _exit_2 = state.getExit();
      _exit_2.setSpecification(exitOp);
      Collection<Transition> _allOutgoings = StateMachineUtil.allOutgoings(state);
      for (final Transition t_1 : _allOutgoings) {
        UMLChain _chain_1 = UMLChain.chain(t_1);
        Behavior _exit_3 = state.getExit();
        _chain_1.prepend(_exit_3);
      }
    }
  }
  
  public String entryFctName(final State state) {
    String _name = state.getName();
    String _plus = ((GlobalConstants.ENTRY_ACTION_FUNC_PREFIX + GlobalConstants.FUNC_NAME_QUAL_NAME_SEP) + _name);
    String _plus_1 = (_plus + GlobalConstants.FUNC_NAME_QUAL_NAME_SEP);
    Behavior _entry = state.getEntry();
    String _name_1 = _entry.getName();
    return (_plus_1 + _name_1);
  }
  
  public String exitFctName(final State state) {
    String _name = state.getName();
    String _plus = ((GlobalConstants.EXIT_ACTION_FUNC_PREFIX + GlobalConstants.FUNC_NAME_QUAL_NAME_SEP) + _name);
    String _plus_1 = (_plus + GlobalConstants.FUNC_NAME_QUAL_NAME_SEP);
    Behavior _exit = state.getExit();
    String _name_1 = _exit.getName();
    return (_plus_1 + _name_1);
  }
  
  /**
   * Creates entry transitions for incoming arrows ending at the composite
   * state's boundary or history pseudo-state. Creates exit transitions
   * implied by group transitions.
   */
  public void phase1ProcessCompositeState(final State state) {
    this.annotateExplicitEntryTransitions(state);
    this.createEntryPoints(state);
    Pseudostate historyChoicePoint = this.replaceHistoryPoint(state);
    this.removeInitialTransition(state, historyChoicePoint);
    this.annotateExplicitExitTransitions(state);
    this.createExitPoints(state);
    this.createExitTransitions(state);
    EList<State> _subStates = StateMachineUtil.subStates(state);
    for (final State subState : _subStates) {
      this.phase1ProcessState(subState);
    }
  }
  
  /**
   * Prepends the state's entry action to each explicit entry transition
   * (transition from an entry point to a sub-state).
   */
  public void annotateExplicitEntryTransitions(final State state) {
    EList<Pseudostate> _entryPoints = StateMachineUtil.entryPoints(state);
    for (final Pseudostate p : _entryPoints) {
      EList<Transition> _outgoings = p.getOutgoings();
      boolean _isEmpty = _outgoings.isEmpty();
      boolean _not = (!_isEmpty);
      if (_not) {
        EList<Transition> _outgoings_1 = p.getOutgoings();
        for (final Transition t : _outgoings_1) {
          UMLChain _chain = UMLChain.chain(t);
          Behavior _entry = state.getEntry();
          _chain.append(_entry);
        }
      }
    }
  }
  
  /**
   * Creates entry points for every direct incoming transition (transition
   * ending directly on the state's boundary).
   */
  public void createEntryPoints(final State state) {
    EList<Transition> _incomings = state.getIncomings();
    for (final Transition t : _incomings) {
      {
        String _newFreshEntryPointName = this.getNewFreshEntryPointName(state);
        final Pseudostate p = state.createConnectionPoint(_newFreshEntryPointName);
        p.setKind(PseudostateKind.ENTRY_POINT_LITERAL);
        t.setTarget(p);
      }
    }
  }
  
  /**
   * Replaces the (deep)history pseudo-state with a choice point with transitions
   * to each sub-state with conditions checking the value of the history table
   * entry for this composite-state.
   */
  public Pseudostate replaceHistoryPoint(final State state) {
    Region _region = StateMachineUtil.region(state);
    final Pseudostate c = StateMachineUtil.deepHistory(_region);
    boolean _notEquals = (!Objects.equal(c, null));
    if (_notEquals) {
      c.setKind(PseudostateKind.CHOICE_LITERAL);
    } else {
      return null;
    }
    EList<State> _subStates = StateMachineUtil.subStates(state);
    for (final State subState : _subStates) {
      {
        Vertex targetVertex = null;
        EList<Region> _regions = subState.getRegions();
        int _size = _regions.size();
        boolean _equals = (_size == 0);
        if (_equals) {
          targetVertex = subState;
        } else {
          Region _region_1 = StateMachineUtil.region(subState);
          Pseudostate _deepHistory = StateMachineUtil.deepHistory(_region_1);
          targetVertex = _deepHistory;
          boolean _equals_1 = Objects.equal(targetVertex, null);
          if (_equals_1) {
            targetVertex = subState;
          }
        }
        Region _region_2 = StateMachineUtil.region(state);
        String _newFreshTransitionName = this.getNewFreshTransitionName(state);
        final Transition newTransition = _region_2.createTransition(_newFreshTransitionName);
        newTransition.setSource(c);
        newTransition.setTarget(targetVertex);
        final Constraint guard = newTransition.createGuard("CheckHistory");
        StringConcatenation _builder = new StringConcatenation();
        _builder.append("history[");
        String _name = state.getName();
        _builder.append(_name, "");
        _builder.append("] == ");
        String _name_1 = subState.getName();
        _builder.append(_name_1, "");
        BehaviorUtil.createOpaqueExpression(guard, _builder.toString());
        Behavior _entry = state.getEntry();
        boolean _notEquals_1 = (!Objects.equal(_entry, null));
        if (_notEquals_1) {
          UMLChain _chain = UMLChain.chain(newTransition);
          Behavior _entry_1 = state.getEntry();
          _chain.append(_entry_1);
        }
      }
    }
    Region _region_1 = StateMachineUtil.region(state);
    final Pseudostate initialState = StateMachineUtil.initialState(_region_1);
    boolean _notEquals_1 = (!Objects.equal(initialState, null));
    if (_notEquals_1) {
      EList<Transition> _outgoings = initialState.getOutgoings();
      final Transition initialTransition = _outgoings.get(0);
      Region _region_2 = StateMachineUtil.region(state);
      String _newFreshTransitionName = this.getNewFreshTransitionName(state);
      final Transition newTransition = _region_2.createTransition(_newFreshTransitionName);
      newTransition.setSource(c);
      Vertex _target = initialTransition.getTarget();
      newTransition.setTarget(_target);
      final Constraint guard = newTransition.createGuard("CheckHistory");
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("history[");
      String _name = state.getName();
      _builder.append(_name, "");
      _builder.append("] == ");
      _builder.append(GlobalConstants.UNVISITED, "");
      BehaviorUtil.createOpaqueExpression(guard, _builder.toString());
      Behavior _effect = initialTransition.getEffect();
      boolean _notEquals_2 = (!Objects.equal(_effect, null));
      if (_notEquals_2) {
        Behavior _effect_1 = initialTransition.getEffect();
        newTransition.setEffect(_effect_1);
        UMLChain _chain = UMLChain.chain(newTransition);
        Behavior _effect_2 = newTransition.getEffect();
        _chain.append(_effect_2);
      }
      UMLChain _chain_1 = UMLChain.chain(newTransition);
      Behavior _entry = state.getEntry();
      _chain_1.append(_entry);
    } else {
      Region _region_3 = StateMachineUtil.region(state);
      String _newFreshName = this.getNewFreshName(state);
      final State border = StateMachineUtil.createState(_region_3, _newFreshName);
      Region _region_4 = StateMachineUtil.region(state);
      String _newFreshTransitionName_1 = this.getNewFreshTransitionName(state);
      final Transition newTransition_1 = _region_4.createTransition(_newFreshTransitionName_1);
      newTransition_1.setSource(c);
      newTransition_1.setTarget(border);
      final Constraint guard_1 = newTransition_1.createGuard("Check History");
      StringConcatenation _builder_1 = new StringConcatenation();
      _builder_1.append("history[");
      String _name_1 = state.getName();
      _builder_1.append(_name_1, "");
      _builder_1.append("] == ");
      _builder_1.append(GlobalConstants.UNVISITED, "");
      BehaviorUtil.createOpaqueExpression(guard_1, _builder_1.toString());
      Behavior _entry_1 = state.getEntry();
      boolean _notEquals_3 = (!Objects.equal(_entry_1, null));
      if (_notEquals_3) {
        UMLChain _chain_2 = UMLChain.chain(newTransition_1);
        Behavior _entry_2 = state.getEntry();
        _chain_2.append(_entry_2);
      }
    }
    return c;
  }
  
  /**
   * Remove the initial pseudo-state and transition and create a transition
   * from each entry point which does not have an explicit entry transition
   * to a sub-state, to the initial state.
   */
  public Boolean removeInitialTransition(final State state, final Pseudostate historyChoicePoint) {
    boolean _xblockexpression = false;
    {
      Region _region = StateMachineUtil.region(state);
      final Pseudostate initialState = StateMachineUtil.initialState(_region);
      boolean _xifexpression = false;
      boolean _notEquals = (!Objects.equal(initialState, null));
      if (_notEquals) {
        boolean _xblockexpression_1 = false;
        {
          EList<Transition> _outgoings = initialState.getOutgoings();
          final Transition initialTransition = _outgoings.get(0);
          EList<Pseudostate> _entryPoints = StateMachineUtil.entryPoints(state);
          for (final Pseudostate p : _entryPoints) {
            EList<Transition> _outgoings_1 = p.getOutgoings();
            boolean _isEmpty = _outgoings_1.isEmpty();
            if (_isEmpty) {
              Region _region_1 = StateMachineUtil.region(state);
              String _newFreshTransitionName = this.getNewFreshTransitionName(state);
              final Transition newTransition = _region_1.createTransition(_newFreshTransitionName);
              newTransition.setSource(p);
              Vertex _target = initialTransition.getTarget();
              newTransition.setTarget(_target);
              UMLChain _chain = UMLChain.chain(newTransition);
              Behavior _entry = state.getEntry();
              _chain.append(_entry);
              UMLChain _chain_1 = UMLChain.chain(newTransition);
              Behavior _effect = initialTransition.getEffect();
              _chain_1.append(_effect);
            }
          }
          Region _region_2 = StateMachineUtil.region(state);
          EList<Transition> _transitions = _region_2.getTransitions();
          _xblockexpression_1 = _transitions.remove(initialTransition);
        }
        _xifexpression = _xblockexpression_1;
      } else {
        EList<Pseudostate> _entryPoints = StateMachineUtil.entryPoints(state);
        for (final Pseudostate p : _entryPoints) {
          EList<Transition> _outgoings = p.getOutgoings();
          boolean _isEmpty = _outgoings.isEmpty();
          if (_isEmpty) {
            Region _region_1 = StateMachineUtil.region(state);
            String _newFreshTransitionName = this.getNewFreshTransitionName(state);
            final Transition newTransition = _region_1.createTransition(_newFreshTransitionName);
            newTransition.setSource(p);
            newTransition.setTarget(historyChoicePoint);
            Behavior _entry = state.getEntry();
            newTransition.setEffect(_entry);
          }
        }
      }
      _xblockexpression = _xifexpression;
    }
    return Boolean.valueOf(_xblockexpression);
  }
  
  /**
   * Annotate explicit exit transitions (transitions from a sub-state to an
   * exit point) with a "history saving action", an action that sets the
   * history table entry for the composite state according to the sub-state
   * that is exited.
   */
  public void annotateExplicitExitTransitions(final State state) {
    EList<Pseudostate> _exitPoints = StateMachineUtil.exitPoints(state);
    for (final Pseudostate p : _exitPoints) {
      EList<Transition> _incomings = p.getIncomings();
      boolean _isEmpty = _incomings.isEmpty();
      boolean _not = (!_isEmpty);
      if (_not) {
        EList<Transition> _incomings_1 = p.getIncomings();
        for (final Transition t : _incomings_1) {
          {
            boolean _equals = Objects.equal(this.saveHistoryBehavior, null);
            if (_equals) {
              this.createSaveHistoryBehavior();
            }
            final UMLAction saveHistoryAction = new UMLAction(this.saveHistoryBehavior);
            String _name = state.getName();
            saveHistoryAction.addParam(_name);
            Vertex _source = t.getSource();
            String _name_1 = _source.getName();
            saveHistoryAction.addParam(_name_1);
            UMLChain _chain = UMLChain.chain(t);
            _chain.append(saveHistoryAction);
            Behavior _exit = state.getExit();
            boolean _notEquals = (!Objects.equal(_exit, null));
            if (_notEquals) {
              UMLChain _chain_1 = UMLChain.chain(t);
              Behavior _exit_1 = state.getExit();
              _chain_1.append(_exit_1);
            }
          }
        }
      }
    }
  }
  
  /**
   * Create exit points for each direct outgoing transition (transition leaving
   * directly from the composite state's boundary).
   */
  public void createExitPoints(final State state) {
    EList<Transition> _outgoings = state.getOutgoings();
    for (final Transition t : _outgoings) {
      {
        String _newFreshExitPointName = this.getNewFreshExitPointName(state);
        final Pseudostate p = state.createConnectionPoint(_newFreshExitPointName);
        p.setKind(PseudostateKind.EXIT_POINT_LITERAL);
        t.setSource(p);
      }
    }
  }
  
  public Behavior createSaveHistoryBehavior() {
    EClass _opaqueBehavior = UMLPackage.eINSTANCE.getOpaqueBehavior();
    Behavior _createOwnedBehavior = this.clazz.createOwnedBehavior(GlobalConstants.SAVE_HISTORY_FUNC_NAME, _opaqueBehavior);
    return this.saveHistoryBehavior = _createOwnedBehavior;
  }
  
  /**
   * Create explicit exit transitions implied by group transitions. For each
   * group transition creates a transition from every sub-state to the
   * group-transitions's source exit point. It annotates these transitions
   * with a "history saving action", an action which stores the sub-state
   * in the history table entry for this composite state.
   */
  public void createExitTransitions(final State state) {
    EList<Pseudostate> _exitPoints = StateMachineUtil.exitPoints(state);
    for (final Pseudostate p : _exitPoints) {
      EList<Transition> _incomings = p.getIncomings();
      boolean _isEmpty = _incomings.isEmpty();
      if (_isEmpty) {
        EList<State> _subStates = StateMachineUtil.subStates(state);
        for (final State subState : _subStates) {
          EList<Transition> _outgoings = p.getOutgoings();
          for (final Transition t : _outgoings) {
            {
              boolean _equals = Objects.equal(this.saveHistoryBehavior, null);
              if (_equals) {
                this.createSaveHistoryBehavior();
              }
              Region _region = StateMachineUtil.region(state);
              String _newFreshTransitionName = this.getNewFreshTransitionName(state);
              final Transition newTransition = _region.createTransition(_newFreshTransitionName);
              newTransition.setSource(subState);
              newTransition.setTarget(p);
              EList<Trigger> _triggers = t.getTriggers();
              for (final Trigger trigger : _triggers) {
                {
                  String _name = trigger.getName();
                  final Trigger newTrigger = newTransition.createTrigger(_name);
                  Event _event = trigger.getEvent();
                  newTrigger.setEvent(_event);
                }
              }
              Constraint _guard = t.getGuard();
              newTransition.setGuard(_guard);
              final UMLAction saveHistoryAction = new UMLAction(this.saveHistoryBehavior);
              String _name = state.getName();
              saveHistoryAction.addParam(_name);
              String _name_1 = subState.getName();
              saveHistoryAction.addParam(_name_1);
              UMLChain _chain = UMLChain.chain(newTransition);
              _chain.append(saveHistoryAction);
              UMLChain _chain_1 = UMLChain.chain(newTransition);
              Behavior _exit = state.getExit();
              _chain_1.append(_exit);
              Behavior _effect = t.getEffect();
              if ((_effect instanceof OpaqueBehavior)) {
                final OpaqueBehavior newBehavior = BehaviorUtil.createOpaqueEffect(newTransition);
                Behavior _effect_1 = t.getEffect();
                String _name_2 = _effect_1.getName();
                newBehavior.setName(_name_2);
                Behavior _effect_2 = t.getEffect();
                String _body = BehaviorUtil.body(_effect_2);
                BehaviorUtil.set(newBehavior, _body);
              }
            }
          }
        }
        EList<Transition> _outgoings_1 = p.getOutgoings();
        for (final Transition t_1 : _outgoings_1) {
          {
            EList<Trigger> _triggers = t_1.getTriggers();
            _triggers.clear();
            t_1.setGuard(null);
            t_1.setEffect(null);
          }
        }
      }
    }
  }
  
  /**
   * Flatten composite states.
   */
  protected StateMachine phase2(final StateMachine m) {
    EList<State> _states = StateMachineUtil.states(m);
    State[] _clone = ((State[])Conversions.unwrapArray(_states, State.class)).clone();
    for (final State s : _clone) {
      {
        this.phase2ProcessState(s);
        EList<State> _subStates = StateMachineUtil.subStates(s);
        boolean _isEmpty = _subStates.isEmpty();
        boolean _not = (!_isEmpty);
        if (_not) {
          Region _region = StateMachineUtil.region(m);
          StateMachineUtil.moveContents(_region, s);
          EList<Region> _regions = s.getRegions();
          _regions.clear();
        }
      }
    }
    return m;
  }
  
  /**
   * Removes entry and exit points of a simple state, making all transitions
   * to and from it, direct.
   */
  protected void _phase2ProcessState(final State state) {
    EList<Region> _regions = state.getRegions();
    boolean _isEmpty = _regions.isEmpty();
    if (_isEmpty) {
      this.phase2ProcessSimpleState(state);
    } else {
      this.phase2ProcessComplexState(state);
    }
  }
  
  /**
   * Removes entry and exit points of a simple state, making all transitions
   * to and from it, direct.
   */
  protected void _phase2ProcessSimpleState(final State state) {
    Collection<Transition> _indirectIncomings = StateMachineUtil.indirectIncomings(state);
    for (final Transition t : _indirectIncomings) {
      t.setTarget(state);
    }
    EList<Transition> _outgoings = state.getOutgoings();
    for (final Transition t_1 : _outgoings) {
      t_1.setSource(state);
    }
    EList<Pseudostate> _entryPoints = StateMachineUtil.entryPoints(state);
    for (final Pseudostate p : _entryPoints) {
      EList<Pseudostate> _connectionPoints = state.getConnectionPoints();
      _connectionPoints.remove(p);
    }
    EList<Pseudostate> _exitPoints = StateMachineUtil.exitPoints(state);
    for (final Pseudostate p_1 : _exitPoints) {
      EList<Pseudostate> _connectionPoints_1 = state.getConnectionPoints();
      _connectionPoints_1.remove(p_1);
    }
  }
  
  /**
   * Replaces all entry and exit points of the composite state with junction
   * points, and recursively apply this to sub-states.
   */
  protected void _phase2ProcessComplexState(final State state) {
    EList<Pseudostate> _connectionPoints = state.getConnectionPoints();
    Pseudostate[] _clone = ((Pseudostate[])Conversions.unwrapArray(_connectionPoints, Pseudostate.class)).clone();
    for (final Pseudostate p : _clone) {
      {
        Region _region = StateMachineUtil.region(state);
        String _name = p.getName();
        String _plus = (GlobalConstants.JUNCTION_FUNC_PREFIX + _name);
        Pseudostate j = StateMachineUtil.createPseudostate(_region, _plus);
        j.setKind(PseudostateKind.JUNCTION_LITERAL);
        EList<Transition> _incomings = p.getIncomings();
        for (final Transition t : _incomings) {
          t.setTarget(j);
        }
        EList<Transition> _outgoings = p.getOutgoings();
        for (final Transition t_1 : _outgoings) {
          t_1.setSource(j);
        }
        PseudostateKind _kind = p.getKind();
        boolean _equals = Objects.equal(_kind, PseudostateKind.ENTRY_POINT_LITERAL);
        if (_equals) {
          EList<Pseudostate> _connectionPoints_1 = state.getConnectionPoints();
          _connectionPoints_1.remove(p);
        } else {
          PseudostateKind _kind_1 = p.getKind();
          boolean _equals_1 = Objects.equal(_kind_1, PseudostateKind.EXIT_POINT_LITERAL);
          if (_equals_1) {
            EList<Pseudostate> _connectionPoints_2 = state.getConnectionPoints();
            _connectionPoints_2.remove(p);
          }
        }
      }
    }
    EList<State> _subStates = StateMachineUtil.subStates(state);
    State[] _clone_1 = ((State[])Conversions.unwrapArray(_subStates, State.class)).clone();
    for (final State subState : _clone_1) {
      {
        this.phase2ProcessState(subState);
        EList<Region> _regions = subState.getRegions();
        boolean _isEmpty = _regions.isEmpty();
        boolean _not = (!_isEmpty);
        if (_not) {
          Region _region = StateMachineUtil.region(state);
          StateMachineUtil.moveContents(_region, subState);
          EList<Region> _regions_1 = subState.getRegions();
          _regions_1.clear();
        }
      }
    }
  }
  
  /**
   * Auxiliary methods
   */
  public String getNewFreshName(final NamedElement parent) {
    this.freshNameCounter++;
    String _namePrefix = this.namePrefix(parent);
    String _plus = (_namePrefix + this.FRESH_NAME_PREFIX);
    return (_plus + Integer.valueOf(this.freshNameCounter));
  }
  
  public String getNewFreshEntryPointName(final NamedElement parent) {
    this.freshEntryPointNameCounter++;
    String _namePrefix = this.namePrefix(parent);
    String _plus = (_namePrefix + this.FRESH_ENTRYPOINT_NAME_PREFIX);
    return (_plus + Integer.valueOf(this.freshEntryPointNameCounter));
  }
  
  public String getNewFreshExitPointName(final NamedElement parent) {
    this.freshExitPointNameCounter++;
    String _namePrefix = this.namePrefix(parent);
    String _plus = (_namePrefix + this.FRESH_EXITPOINT_NAME_PREFIX);
    return (_plus + Integer.valueOf(this.freshExitPointNameCounter));
  }
  
  public String getNewFreshChoicePointName(final NamedElement parent) {
    this.freshChoicePointNameCounter++;
    String _namePrefix = this.namePrefix(parent);
    String _plus = (_namePrefix + this.FRESH_CHOICEPOINT_NAME_PREFIX);
    return (_plus + Integer.valueOf(this.freshChoicePointNameCounter));
  }
  
  public String getNewFreshJunctionPointName(final NamedElement parent) {
    this.freshJunctionPointNameCounter++;
    String _namePrefix = this.namePrefix(parent);
    String _plus = (_namePrefix + this.FRESH_JUNCTIONPOINT_NAME_PREFIX);
    return (_plus + Integer.valueOf(this.freshJunctionPointNameCounter));
  }
  
  public String getNewFreshTransitionName(final NamedElement parent) {
    this.freshTransitionNameCounter++;
    String _namePrefix = this.namePrefix(parent);
    String _plus = (_namePrefix + this.FRESH_TRANSITION_NAME_PREFIX);
    return (_plus + Integer.valueOf(this.freshTransitionNameCounter));
  }
  
  public String namePrefix(final NamedElement ne) {
    String _name = ne.getName();
    return (_name + "_");
  }
  
  public void phase2ProcessState(final State state) {
    _phase2ProcessState(state);
    return;
  }
  
  public void phase2ProcessSimpleState(final State state) {
    _phase2ProcessSimpleState(state);
    return;
  }
  
  public void phase2ProcessComplexState(final State state) {
    _phase2ProcessComplexState(state);
    return;
  }
}
